unit Sub;

interface

uses Windows;

{ tyyppi PBinFile kytetn LoadBinaryFile- ja FreeBinaryFile -funktioissa }
type
  PBinFile = ^TBinFile;
  TBinFile = packed record
    Pit: Longint;
    Tbl: array[0..0] of Byte;
  end;

type
 TByteRec = record
   Lo, Hi: Byte;
 end;

{ compares A with B, length Pit bytes, True if equal }
function Equal(const A, B; Pit: Integer): Boolean; Register;
{ register tarkoittaa, ett parametrit tulee assemblerin rekistereihin,
  se om toki oletuksenakin Delphiss }

{ Onko A nolla, pituus Pit byte }
function Nolla(const A; Pit: Integer): Boolean; Register;

{ Kuten StrToInt, mutta hyvksyy mys tyhjn jonon }
function StringToInt(const S: String): Integer;

{ hakee seuraavan luvun IResult merkkijonon S indeksist Ix lhtien
  ja askeltaa Ix:n seuraavan non-numeromerkin kohtaan.
  Return True, jos lytyi }
function StrToNextInt(const S: String; var Ix, IResult: Integer): Boolean;

{ kuten edell, mutta hakee reaaliluvun tyyppi Single (tarvitsee 4 tavua) }
function StrToNextVal(const S: String; var Ix: Integer; var IResult: Single): Boolean;

{ hakee merkkijonosta S seuraavan sanan. Pilkku erotusmerkkin }
function StrToNextStr(const S: String; var Ix: Integer): String;

{ Search Substring from String, beginning from start position i.
  Not case sensitive. Result equal as Pos function. }
function LStrPosIC(const substr, s : AnsiString; i: Integer) : Integer; Register;

{ same as above but backwards }
function LStrBackIC(const substr, s : AnsiString; i: Integer) : Integer; Register;

{ Read binary file to memory and allocates neccessary memory space. }
function LoadBinaryFile(const Name: String): PBinFile;

{ Free BinFile type if not nil }
procedure FreeBinaryFile(var P: PBinFile);

{ Abs-funktio mys word-luvuista }
function Its(Luku: Integer): Integer;

{ hakee pienimmn 1-bitin numeron }
function SeekLeastBit(W: Integer): Integer; Register;

{ hakee suurimman 1-bitin numeron }
function SeekMostBit(W: Integer): Integer; Register;

{ muuntaa bittinumeron 0..31 maskiksi }
function BitToMask(W: Integer): Integer; Register;

{ onko > 1 bitti pll }
function BitMulti(W: Integer): Boolean; Register;

{ montako 1-bitti }
function BitKpl(W: Integer): Integer; Register;

{ muuntaa bittimerkin bittinumeroksi }
function HexToBit(C: Char): Integer;

{ Signed divide W by B^2 }
function ShiftAR(W, B: Integer): Integer; Register;

procedure SwapIntegers(var I1, I2: Integer);

{ mik on bitin Pos tila W:ss }
function BitStatus(W, Pos: Integer): Boolean; Register;

function IntToBinStr(W, L: Integer): String;

function IntPower(Base: Extended; Exponent: Integer): Extended; Register;

function Log10(X: Extended): Extended;

{ Korvaa merkkijonossa S jokaisen S1:ss esiintyvn merkin
  vastaavalla S2-merkkijonon merkill.
  Kytetn esim. scandinaavisten merkkien muunnoksissa}
procedure ReplaceChrSets(const S1, S2: AnsiString; var S: AnsiString); Register;

procedure FixScandString(var S: AnsiString);

{ Tiedostonimest nimi ilman extensiota }
function FileNameOnly(const FileName: string): string;

{ asettaa I1:ksi suuremman I1 tai I2- luvuista }
procedure SetMaxInt(var I1: Integer; I2: Integer);

{ asettaa I1:ksi pienemmn I1 tai I2- luvuista }
procedure SetMinInt(var I1: Integer; I2: Integer);

{ onko (X,Y) R:n sisll }
function XYInRect(X, Y: Integer; const R: TRect): Boolean;

{ leikkaako R1 ja R2 toisiaan }
function IsIntersectRect(const R1, R2: TRect): Boolean;

function JonoToStr(MaxPit: Integer; var Jono): String;

const
    DosScandString = #$F8#$84#$94#$8F#$86#$8E#$99;
    WinScandString = '';
    DosScandString0 = #$F8#$84#$94#$8F#$86#$8E#$99#0;
    WinScandString0 = '?';

HexChr: array[0..15] of Char = '0123456789ABCDEF';

implementation

uses SysUtils;

(* { compares A with B, length Pit bytes, True if equal }
function Equal(const A, B; Pit: Integer): Boolean;
var
  i: Integer;
begin
  Result := False;
  if Pit > 0 then
  for i := 0 to Pit - 1 do
  if TByteArray(A)[i] <> TByteArray(B)[i] then Exit;
  Result := True;
end; *)

{ compares A with B, length Pit bytes, True if equal }
function Equal(const A, B; Pit: Integer): Boolean;
{     ->EAX     Pointer to A  }
{     ->EDX     Pointer to B  }
{     ->ECX     Pit           }
asm
        PUSH    EDI
        PUSH    ESI
        MOV     EDI,EDX
        MOV     ESI,EAX
        XOR     EAX,EAX
        OR      ECX,ECX
        JZ      @@1
        REPE    CMPSB
        JNE     @@2
@@1:    INC     EAX
@@2:    POP     ESI
        POP     EDI
end;

{ Onko A nolla, pituus Pit byte }
function Nolla(const A; Pit: Integer): Boolean;
{     ->EAX     Pointer to A  }
{     ->EDX     Pit           }
asm
        PUSH    EDI
        MOV     EDI,EAX
        MOV     ECX,EDX
        XOR     EAX,EAX
        OR      ECX,ECX
        JZ      @@1
        REPE    SCASB
        JNE     @@2
@@1:    INC     EAX
@@2:    POP     EDI
end;

function StringToInt(const S: String): Integer;
var
  ApuI: Integer;
begin
  Val(S, Result, ApuI);
  if ApuI <> 0 then Result := 0;
end;

function StrToNextInt(const S: String; var Ix, IResult: Integer): Boolean;
var
  ApuI: Integer;
begin
  while (Ix <= Length(S)) and not (S[Ix] in ['0'..'9']) do Inc(Ix);
  ApuI := Ix;
  while (Ix <= Length(S)) and (S[Ix] in ['0'..'9']) do Inc(Ix);
  if (ApuI <= Length(S)) and (Ix > ApuI) then
  begin
    Val(Copy(S, ApuI, Ix - ApuI), IResult, ApuI);
    Result := ApuI = 0;
  end
  else Result := False;
end;

function StrToNextStr(const S: String; var Ix: Integer): String;
var
  ApuI, ApuH: Integer;
begin
  Result := '';
  if Ix > Length(S) then Exit;
  while (Ix <= Length(S)) and (S[Ix] = ' ') do Inc(Ix);
  ApuI := LStrPosIC(',', S, Ix);
  if ApuI = 0 then ApuI := Length(S) + 1;
  if (ApuI > Ix) then
  begin
    ApuH := ApuI - 1;
    while (ApuH > Ix) and (S[ApuH] = ' ') do Dec(ApuH);
    Result := Copy(S, Ix, ApuH - Ix + 1);
  end;
  Ix := ApuI + 1;
end;

function StrToNextVal(const S: String; var Ix: Integer; var IResult: Single): Boolean;
var
  ApuI: Integer;
begin
  while (Ix <= Length(S)) and not (S[Ix] in ['-', '.', '0'..'9']) do Inc(Ix);
  ApuI := Ix;
  while (Ix <= Length(S)) and (S[Ix] in ['-', '.', '0'..'9']) do Inc(Ix);
  if (ApuI <= Length(S)) and (Ix > ApuI) then
  begin
    Val(Copy(S, ApuI, Ix - ApuI), IResult, ApuI);
    Result := ApuI = 0;
  end
  else Result := False;
end;

type
  StrRec = packed record
    allocSiz: Longint;
    refCnt: Longint;
    length: Longint;
  end;

const
        skew = sizeof(StrRec);
        rOff = sizeof(StrRec) - sizeof(Longint); { refCnt offset }
        overHead = sizeof(StrRec) + 1;


{ Search Substring from String, beginning from start position i
  Tm on muunnettu LStrPos-functiosta (Pos-funktion long string -versio)}
function LStrPosIC(const substr, s : AnsiString; i: Integer) : Integer;
asm
{     ->EAX     Pointer to substr               }
{       EDX     Pointer to string               }
{     ->ECX     StartPosition to string         }
{     <-EAX     Position of substr in s or 0    }

        TEST    EAX,EAX
        JE      @@noWork

        TEST    EDX,EDX
        JE      @@fail2

        TEST    ECX,ECX
        JE      @@fail2

        PUSH    EBX
        PUSH    ESI
        PUSH    EDI

        MOV     EDI,EAX      { Point EDI to substr           }
        MOV     ESI,EDX      { Point ESI to s                }
        MOV     EAX,ECX      { start position }

        MOV     ECX,[ESI-skew].StrRec.length    { ECX = Length(s)               }

        INC     ECX
        SUB     ECX,EAX      { - start position }
        JBE     @@fail3

        PUSH    ESI          { remember s position to calculate index }
        DEC     ESI
        ADD     ESI,EAX      { Point ESI to first char of s  }

        MOV     EDX,[EDI-skew].StrRec.length    { EDX = Length(substr) }

        DEC     EDX          { EDX = Length(substr) - 1              }
        JS      @@fail       { < 0 ? return 0                        }
        MOV     AH,[EDI]     { AL = first char of substr             }
        INC     EDI          { Point EDI to 2'nd char of substr      }

        SUB     ECX,EDX      { #positions in s to look at    }
                             { = Length(s) - Length(substr) + 1      }
        JLE     @@fail
        CLD                  { direct = inc }
@@loop:
        LODSB
        XOR     AL,AH
        AND     AL,0DFH      { not case sensitive }
        LOOPNE  @@loop
                             { REPNE   SCASB }
        JNE     @@fail
        TEST    EDX,EDX
        JZ      @@found      { no more characters }
        MOV     EBX,ECX      { save outer loop counter               }
        PUSH    EDI          { save outer loop substr pointer        }
        PUSH    ESI          { save outer loop s pointer             }

        MOV     ECX,EDX
        PUSH    EAX
@@loop2:
        MOV     AH,[EDI]
        INC     EDI
        LODSB
        XOR     AL,AH
        AND     AL,0DFH       { not case sensitive }
        LOOPE   @@loop2
        POP     EAX
                              { REPE    CMPSB  }
        POP     ESI           { restore outer loop s pointer  }
        POP     EDI           { restore outer loop substr pointer     }
        JE      @@found
        MOV     ECX,EBX       { restore outer loop counter    }
        JMP     @@loop

@@fail:
        POP     EDX           { get rid of saved s pointer    }
@@fail3:
        XOR     EAX,EAX
        JMP     @@exit

@@fail2:
        XOR     EAX,EAX
        JMP     @@noWork

@@found:
        POP     EDX           { restore pointer to first char of s    }
        MOV     EAX,ESI       { ESI points of char after match        }
        SUB     EAX,EDX       { the difference is the correct index   }
@@exit:
        POP     EDI
        POP     ESI
        POP     EBX
@@noWork:
end;

{ kuten LStrPosIC, mutta hakee takaperin }
function LStrBackIC(const substr, s : AnsiString; i: Integer) : Integer;
asm
{     ->EAX     Pointer to substr               }
{       EDX     Pointer to string               }
{     ->ECX     StartPosition to string         }
{     <-EAX     Position of substr in s or 0    }

        TEST    EAX,EAX
        JE      @@noWork

        TEST    EDX,EDX
        JE      @@fail2

        TEST    ECX,ECX
        JE      @@fail2

        PUSH    EBX
        PUSH    ESI
        PUSH    EDI

        MOV     EDI,EAX      { Point EDI to substr           }
        MOV     ESI,EDX      { Point ESI to s                }

        PUSH    ESI          { remember s position to calculate index }

        MOV     EDX,[EDI-skew].StrRec.length    { EDX = Length(substr) }

        { check that search not goes above s limit }
        MOV     EAX,ECX      { start pos }
        DEC     EAX
        ADD     EAX,EDX      { + length(substr) }
        CMP     EAX,[ESI-skew].StrRec.length    { gt than Length(s) }
        JBE     @@cont
        { above the length, let's limit }
        MOV     EAX,[ESI-skew].StrRec.length    { Length(s) }
        SUB     EAX,EDX      { - length(subStr }
        JB      @@fail
        INC     EAX
        { correct ECX }
        MOV     ECX,EAX
        { continue }
@@cont: DEC     ESI
        ADD     ESI,ECX      { Point ESI to first char of s  }
        DEC     EDX          { EDX = Length(substr) - 1              }
        JS      @@fail       { < 0 ? return 0                        }
        MOV     AH,[EDI]     { AL = first char of substr             }
        INC     EDI          { Point EDI to 2'nd char of substr      }

@@loops:
        STD                  { direct flag = dec }
@@loop:
        LODSB
        XOR     AL,AH
        AND     AL,0DFH      { not case sensitive }
        LOOPNE  @@loop
                             { REPNE   SCASB }
        JNE     @@fail
        TEST    EDX,EDX
        JZ      @@found      { no more characters }
        MOV     EBX,ECX      { save outer loop counter               }
        PUSH    EDI          { save outer loop substr pointer        }
        PUSH    ESI          { save outer loop s pointer             }

        MOV     ECX,EDX
        PUSH    EAX
        CLD                  { direction flag = inc }
@@loop2:
        MOV     AH,[EDI]
        INC     EDI
        LODSB
        XOR     AL,AH
        AND     AL,0DFH       { not case sensitive }
        LOOPE   @@loop2
        POP     EAX
                              { REPE    CMPSB  }
        POP     ESI           { restore outer loop s pointer  }
        POP     EDI           { restore outer loop substr pointer     }
        JE      @@found
        MOV     ECX,EBX       { restore outer loop counter    }
        JMP     @@loops

@@fail:
        POP     EDX           { get rid of saved s pointer    }
@@fail3:
        XOR     EAX,EAX
        JMP     @@exit

@@fail2:
        XOR     EAX,EAX
        JMP     @@noWork

@@found:
        POP     EDX           { restore pointer to first char of s    }
        MOV     EAX,ESI       { ESI points of char after match        }
        SUB     EAX,EDX       { the difference   }
        INC     EAX
        INC     EAX           { correct index }
@@exit:
        POP     EDI
        POP     ESI
        POP     EBX
@@noWork:
end;


{ Read binary file to memory.
  Allocates neccessary memory space.
  Result^ record format:
    Length: DWORD;
    Binary: array[0..Length-1] of Byte; }
function LoadBinaryFile(const Name: String): PBinFile;
var
  F: file;
  FLength: Longint;
begin
  Result := nil;
  AssignFile(F, Name);
  {$I-}
  Reset(F, 1);
  if IOResult = 0 then
  begin
    FLength := FileSize(F);
    if IOResult = 0 then
    begin
      Result := AllocMem(SizeOf(Longint) + FLength);
      if Result <> nil then
      begin
        Result^.Pit := FLength;
        BlockRead(F, Result^.Tbl, Result^.Pit, FLength);
        if Result^.Pit <> FLength
        then ReAllocMem(Result, 0);
      end;
    end;
    CloseFile(F);
  end;
  {$I+}
end;

procedure FreeBinaryFile(var P: PBinFile);
begin
  if P <> nil then
  begin
    FreeMem(P, SizeOf(Longint) + P^.Pit);
    P := nil;
  end;
end;

{ Abs-funktio mys word-luvuista }
function Its(Luku: Integer): Integer;
begin
  Its := Abs(Luku);
end;

function SeekLeastBit(W: Integer): Integer;
{ -> EAX = W
  EAX -> Result }
asm
  BSF EAX,EAX
end;

function SeekMostBit(W: Integer): Integer;
{ -> EAX = W
  EAX -> Result }
asm
  BSR EAX,EAX
end;

function BitToMask(W: Integer): Integer;
{ -> EAX = W
  EAX -> Result }
asm
  XOR EDX,EDX
  BTS EDX,EAX
  MOV EAX,EDX
end;

function BitMulti(W: Integer): Boolean;
{ -> EAX = W
  EAX -> Result }
asm
  BSF EDX,EAX
  JZ  @1
  BTR EAX,EDX
  OR  EAX,EAX
  JZ  @1
  MOV EAX,1
@1:
end;

function BitKpl(W: Integer): Integer;
{ -> EAX = W
  EAX -> Result }
asm
  MOV ECX,32
@1:
  BSF EDX,EAX
  JZ  @3
  BTR EAX,EDX
  LOOP @1
@3:
  MOV EAX,32
  SUB EAX,ECX
end;

function HexToBit(C: Char): Integer;
begin
  case C of
    '0'..'9':           Result := Ord(C) - Ord('0');
    'A'..'F', 'a'..'f': Result := (Ord(C) and 7) + 9;
    else                Result := 0;
  end;
end;

{ Signed divide }
function ShiftAR(W, B: Integer): Integer;
{ -> EAX = W
  -> EDX = B
  EAX -> Result }
asm
  MOV ECX,EDX
  SAR EAX,CL
end;

procedure SwapIntegers(var I1, I2: Integer);
var
  I3: Integer;
begin
  I3 := I1;
  I1 := I2;
  I2 := I3;
end;

{ mik on bitin Pos tila W:ss }
function BitStatus(W, Pos: Integer): Boolean;
{ -> EAX = W
  -> EDX = Pos
  EAX -> Result }
asm
  BT EAX,EDX
  MOV EAX,0
  JNC  @1
  INC EAX
@1:
end;

function IntToBinStr(W, L: Integer): String;
var
  Li: Integer;
begin
  SetLength(Result, L);
  if L > 0 then
  for Li := L downto 1 do
    if BitStatus(W, L - Li) then Result[Li] := '1' else Result[Li] := '0';
end;

function IntPower(Base: Extended; Exponent: Integer): Extended;
asm
        mov     ecx, eax
        cdq
        fld1                      { Result := 1 }
        xor     eax, edx
        sub     eax, edx          { eax := Abs(Exponent) }
        jz      @@3
        fld     Base
        jmp     @@2
@@1:    fmul    ST, ST            { X := Base * Base }
@@2:    shr     eax,1
        jnc     @@1
        fmul    ST(1),ST          { Result := Result * X }
        jnz     @@1
        fstp    st                { pop X from FPU stack }
        cmp     ecx, 0
        jge     @@3
        fld1
        fdivrp                    { Result := 1 / Result }
@@3:
        fwait
end;

function Log10(X: Extended): Extended;
  { Log.10(X) := Log.2(X) * Log.10(2) }
asm
        FLDLG2     { Log base ten of 2 }
        FLD     X
        FYL2X
        FWAIT
end;

{ Korvaa merkkijonossa S jokaisen S1:ss esiintyvn merkin
  vastaavalla S2-merkkijonon merkill.
  Kytetn esim. scandinaavisten merkkien muunnoksissa}
procedure ReplaceChrSets(const S1, S2: AnsiString; var S: AnsiString);
asm
{     ->EAX     Pointer to S1         }
{     ->EDX     Pointer to S2         }
{     ->ECX     Pointer to S          }

        TEST    EAX,EAX
        JE      @@loppu

        TEST    EDX,EDX
        JE      @@loppu

        MOV     ECX,[ECX]
        TEST    ECX,ECX
        JE      @@loppu

        PUSH    EBX
        PUSH    ESI
        PUSH    EDI

        MOV     EDI,ECX      { Point EDI to S  }
        MOV     ESI,EAX      { Point ESI to S1 }
        MOV     EBX,EDX      { Point EBX to S2 }

        MOV     ECX,[EDI-skew].StrRec.length    { ECX = Length(S)               }
        TEST    ECX,ECX
        JE      @@loppu2

        MOV     EDX,[ESI-skew].StrRec.length    { EDX = Length(S1) }

        TEST    EDX,EDX
        JE      @@loppu2

        TEST    EDX,[EBX-skew].StrRec.length    { Compare Length(S1) to Length(S2) }
        JL      @@loppu2

        CLD
        JMP     @@loop2

@@loop1:
        POP     EDI          { S }
        POP     ECX          { Length(S) }
        DEC     EDX
        JZ      @@loppu2
        INC     EBX

@@loop2:
        PUSH    ECX
        PUSH    EDI

        LODSB                { AL = [ESI], INC(ESI)  }
@@loop3:
        REPNE   SCASB

        JNE     @@loop1

        MOV     AH,[EBX]
        MOV     [EDI-1],AH

        TEST    ECX,ECX
        JNZ     @@loop3
        JMP     @@loop1
        POP     EDI
        POP     ECX

@@loppu2:
        POP     EDI
        POP     ESI
        POP     EBX
@@loppu:
end;

procedure FixScandString(var S: AnsiString);
begin
  ReplaceChrSets(DosScandString, WinScandString, S);
end;

function FileNameOnly(const FileName: string): string;
var
  ApuI: Integer;
begin
  Result := ExtractFileName(FileName);
  ApuI := Pos('.', Result);
  if ApuI > 0 then SetLength(Result, ApuI - 1);
end;

{ asettaa I1:ksi suuremman I1 tai I2- luvuista }
procedure SetMaxInt(var I1: Integer; I2: Integer);
begin
  if I2 > I1 then I1 := I2;
end;

{ asettaa I1:ksi pienemmn I1 tai I2- luvuista }
procedure SetMinInt(var I1: Integer; I2: Integer);
begin
  if I2 < I1 then I1 := I2;
end;

{ onko (X,Y) R:n sisll }
function XYInRect(X, Y: Integer; const R: TRect): Boolean;
begin
  Result := (X >= R.Left)
        and (X <= R.Right)
        and (Y >= R.Top)
        and (Y <= R.Bottom);
end;

{ leikkaako R1 ja R2 toisiaan }
function IsIntersectRect(const R1, R2: TRect): Boolean;
begin
  Result := (R1.Left <= R2.Right)
        and (R1.Right >= R2.Left)
        and (R1.Top <= R2.Bottom)
        and (R1.Bottom >= R2.Top);
end;

function JonoToStr(MaxPit: Integer; var Jono): String;
var
  Ix: Integer;
begin
  Result := '';
  if MaxPit > 0 then
  for Ix := 0 to MaxPit - 1 do
  if TByteArray(Jono)[Ix] in [$21..$7F]
  then Result := Result + Char(TByteArray(Jono)[Ix]);
end;

end.

